import numpy as np

"""
This function is used along with resizing images from RGB and thermal camera
During the resizing, some gaps might appear, where some pixels are not given a new
value and therefore have a value of zero. This function finds these pixels and
replaces them with the average of their non-zero neighbors.
"""

def remove_zeros(matrix_2D):
    matrix_2D_copy = np.copy(matrix_2D)
    for i in range(matrix_2D.shape[0]):
        for j in range(matrix_2D.shape[1]):
            if matrix_2D_copy[i,j] == 0:
                matrix_2D_copy[i,j] = avg_neighbors(matrix_2D, i, j)
            else:
                matrix_2D_copy[i,j] = matrix_2D[i,j]
    return matrix_2D_copy

"""
Calculate the average of all non-zero neighbors
"""

def avg_neighbors(matrix_2D, row, col):
    avg = 0
    #Upper left
    if row == 0 and col == 0:
        temp_vals = np.array([matrix_2D[0, 1], matrix_2D[1, 1], matrix_2D[1, 0]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)
    #Upper right
    elif row == 0 and col == matrix_2D.shape[1]-1:
        temp_vals = np.array([matrix_2D[1, col], matrix_2D[1, col-1], matrix_2D[0, col-1]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)
    #Lower left
    elif row == matrix_2D.shape[0]-1 and col == 0:
        temp_vals = np.array([matrix_2D[row-1, 0], matrix_2D[row-1, 1], matrix_2D[row, 1]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)
    #Lower right
    elif row == matrix_2D.shape[0]-1 and col == matrix_2D.shape[1]-1:
        temp_vals = np.array([matrix_2D[row, col-1], matrix_2D[row-1, col-1], matrix_2D[row-1, col]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)
    #Leftmost column
    elif col == 0:
        temp_vals = np.array([matrix_2D[row-1, col], matrix_2D[row-1, col+1], matrix_2D[row, col+1], matrix_2D[row+1, col+1], matrix_2D[row+1, col]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)
    #Rightmost column
    elif col == matrix_2D.shape[1]-1:
        temp_vals = np.array([matrix_2D[row-1, col], matrix_2D[row-1, col-1], matrix_2D[row, col-1], matrix_2D[row+1, col-1], matrix_2D[row+1, col]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)
    #Upper row
    elif row == 0:
        temp_vals = np.array([matrix_2D[row, col+1], matrix_2D[row+1, col+1], matrix_2D[row+1, col], matrix_2D[row+1, col-1], matrix_2D[row, col-1]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)
    #Bottom row
    elif row == matrix_2D.shape[0]-1:
        temp_vals = np.array([matrix_2D[row, col+1], matrix_2D[row-1, col+1], matrix_2D[row-1, col], matrix_2D[row-1, col-1], matrix_2D[row, col-1]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)
    else:
        temp_vals = np.array([matrix_2D[row, col+1], matrix_2D[row-1, col+1], matrix_2D[row-1, col], matrix_2D[row-1, col-1], matrix_2D[row, col-1], matrix_2D[row+1, col-1], matrix_2D[row+1, col], matrix_2D[row+1, col+1]])
        temp_vals = temp_vals[temp_vals > 0]
        avg = np.mean(temp_vals)

    return avg



import numpy as np

def repair_bad_col(cube):
    #This function replaces two column of defected pixels on the 10_10_200_22 camera. The column indices are hard coded, and requires the image to have been rotated right side up
    #as well as still be at full size. 

     #Firstly recognize whether there are more than one image in the 'cube'       
    if len(cube.shape) == 2:
        cube = cube[:,:,np.newaxis]
        multilayer = False
    else:
        multilayer = True

    #Define indexes of all bad pixels for first column
    all_row_idx1 = np.arange(0,cube.shape[0])
    good_row_idx1 = all_row_idx1[::2]
    #... and second column
    all_row_idx2 = np.arange(950,cube.shape[0])
    good_row_idx2 = all_row_idx2[::2]
    
    #Iterate through each layer of the cube and repair
    for i in range(cube.shape[2]):
        #First column
        good_row_vals = cube[good_row_idx1,504, i]
        cube[all_row_idx1,504, i] = np.interp(all_row_idx1, good_row_idx1, good_row_vals)
        #Second column
        good_row_vals = cube[good_row_idx2,505, i]
        cube[all_row_idx2,505, i] = np.interp(all_row_idx2, good_row_idx2, good_row_vals)
    #Return input with the same shape as input
    if not multilayer:
        cube = np.squeeze(cube, 2)
    return cube

def repair_nans(cube, max_it = 100):
    #This function replaces all NaN-containing entries in a 2D numpy array and replaces them with the median value of their apparent neighborhood

    #Firstly recognize whether there are more than one image in the 'cube'
    if len(cube.shape) == 2:
        cube = cube[:,:,np.newaxis]
        multilayer = False
    else:
        multilayer = True

    # Remove NaNs from each layer in cube
    for i in range(cube.shape[2]):
        offset = 1 #The offset defines the neighborhood around the 'bad' pixel, which will be used to correct it. An offset of 1 results in a 3x3 neighborhood with the bad pixel in the center
        layer = cube[:,:,i]
        while np.isnan(layer).any(): #As long as there are NaNs do:
            [nan_row, nan_col] = np.where(np.isnan(layer)) #Get indices of bad pixels
            #If there are any edge cases, the entire image is padded with the median value, and the row and column indices are adjusted accordingly
            if np.isin(nan_row, [0,layer.shape[0]-1]).any() or np.isin(nan_col, [0,layer.shape[1]-1]).any(): 
                layer = np.pad(layer, offset, 'constant', constant_values = np.nanmedian(layer))
                nan_row += offset
                nan_col += offset
            for j in range(len(nan_row)): #Loop through every index and replace with the median of the neighborhood
                neighborhood = layer[nan_row[j]-offset:nan_row[j]+offset+1,nan_col[j]-offset:nan_col[j]+offset+1]
                layer[nan_row[j], nan_col[j]] = np.nanmedian(neighborhood) #Replace neighborhood
            offset += 1 #Offset is increased in case there are still bad pixels
            #If max number of iterations is reached, break out of loop - only here to prevent hang-ups
            if offset > max_it:
                print('Break')
                break
        #If edge cases are detected, only the original part of the image is kept - not the padding
        if layer.shape[0] != cube.shape[0]:
            layer = layer[offset - 1:-(offset - 1), offset - 1:-(offset - 1)]
        
        cube[:,:,i] = layer

    #The third dimension of the cube is removed if there is only a single layer
    if not multilayer:
        cube = np.squeeze(cube, 2)
    return cube


def remove_stuck_px(data_cube):
    temp_data_cube = np.copy(data_cube)
    min_img = np.min(data_cube, axis=2)
    zero_idx = np.where(min_img == 0)
    zero_std_idx = np.where(np.std(data_cube, axis = 2) == 0)
    for i in range(len(zero_idx[0])):
        if zero_idx[0][i] == 0 and zero_idx[1][i] == 0:  # upper left
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :]) / 3
        elif zero_idx[0][i] == 0 and zero_idx[1][i] == data_cube.shape[1]-1:  # upper right
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 3
        elif zero_idx[0][i] == data_cube.shape[0]-1 and zero_idx[1][i] == 0:  # lower left
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] - 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :]) / 3
        elif zero_idx[0][i] == data_cube.shape[0]-1 and zero_idx[1][i] == data_cube.shape[1]-1:  # lower right
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] - 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 3
        elif zero_idx[0][i] == 0:  # upper row
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 5
        elif zero_idx[0][i] == data_cube.shape[0]-1:  # lower row
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] - 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 5
        elif zero_idx[1][i] == 0:  # leftmost column
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :]) / 5
        elif zero_idx[1][i] == data_cube.shape[1]-1:  # rightmost column
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] - 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 5
        else:
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 8

    for i in range(len(zero_std_idx[0])):
        if zero_std_idx[0][i] == 0 and zero_std_idx[1][i] == 0:  # upper left
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :]) / 3
        elif zero_std_idx[0][i] == 0 and zero_std_idx[1][i] == data_cube.shape[1]-1:  # upper right
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 3
        elif zero_std_idx[0][i] == data_cube.shape[0]-1 and zero_std_idx[1][i] == 0:  # lower left
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :]) / 3
        elif zero_std_idx[0][i] == data_cube.shape[0]-1 and zero_std_idx[1][i] == data_cube.shape[1]-1:  # lower right
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 3
        elif zero_std_idx[0][i] == 0:  # upper row
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 5
        elif zero_std_idx[0][i] == data_cube.shape[0]-1:  # lower row
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 5
        elif zero_std_idx[1][i] == 0:  # leftmost column
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :]) / 5
        elif zero_std_idx[1][i] == data_cube.shape[1]-1:  # rightmost column
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 5
        else:
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 8

    return temp_data_cube


def remove_outlying_px(data_cube, cut_off):
    temp_data_cube = np.copy(data_cube)
    gray_img = np.mean(temp_data_cube, axis=2)
    gray_img += abs(np.min(gray_img))
    gray_img = gray_img / np.max(gray_img)

    neighbour_img = np.zeros([gray_img.shape[0], gray_img.shape[1]])
    for i in range(gray_img.shape[0]):
        for j in range(gray_img.shape[1]):
            if i == 0 and j == 0:  # upper left
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j + 1] + gray_img[i + 1, j] + gray_img[i + 1, j + 1]) / 3
            elif i == 0 and j == gray_img.shape[1] - 1:  # upper right
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j - 1] + gray_img[i + 1, j] + gray_img[i + 1, j - 1]) / 3
            elif i == gray_img.shape[0] - 1 and j == 0:  # bottom left
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i - 1, j] + gray_img[i, j + 1] + gray_img[i - 1, j + 1]) / 3
            elif i == gray_img.shape[0] - 1 and j == gray_img.shape[1] - 1:  # bottom right
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j - 1] + gray_img[i - 1, j] + gray_img[i - 1, j - 1]) / 3
            elif i == 0:  # top row
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j - 1] + gray_img[i, j + 1] + gray_img[i + 1, j] + gray_img[i + 1, j - 1] + gray_img[i + 1, j + 1]) / 5
            elif i == gray_img.shape[0] - 1:  # bottom row
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j - 1] + gray_img[i, j + 1] + gray_img[i - 1, j] + gray_img[i - 1, j - 1] + gray_img[i - 1, j + 1]) / 5
            elif j == 0:  # leftmost column
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i + 1, j] + gray_img[i - 1, j] + gray_img[i, j + 1] + gray_img[i + 1, j + 1] + gray_img[i - 1, j + 1]) / 5
            elif j == gray_img.shape[1] - 1:  # rightmost column
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i + 1, j] + gray_img[i - 1, j] + gray_img[i, j - 1] + gray_img[i + 1, j - 1] + gray_img[i - 1, j - 1]) / 5
            else:
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i - 1, j - 1] + gray_img[i - 1, j] + gray_img[i - 1, j + 1] + gray_img[i, j + 1] + gray_img[i + 1, j + 1] + gray_img[i + 1, j] + gray_img[i + 1, j - 1] + gray_img[i, j - 1]) / 8

    outlier_idx = np.where(abs(neighbour_img) > cut_off)

    for i in range(len(outlier_idx[0])):
        if outlier_idx[0][i] == 0 and outlier_idx[1][i] == 0:  # upper left
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :]) / 3
        elif outlier_idx[0][i] == 0 and outlier_idx[1][i] == data_cube.shape[1]-1:  # upper right
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 3
        elif outlier_idx[0][i] == data_cube.shape[0]-1 and outlier_idx[1][i] == 0:  # lower left
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :]) / 3
        elif outlier_idx[0][i] == data_cube.shape[0]-1 and outlier_idx[1][i] == data_cube.shape[1]-1:  # lower right
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 3
        elif outlier_idx[0][i] == 0:  # upper row
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 5
        elif outlier_idx[0][i] == data_cube.shape[0]-1:  # lower row
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 5
        elif outlier_idx[1][i] == 0:  # leftmost column
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :]) / 5
        elif outlier_idx[1][i] == data_cube.shape[1]-1:  # rightmost column
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 5
        else:
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 8

    return temp_data_cube, outlier_idx


import numpy as np
import os

#This function imports the output file into one large 2D array. The first three columns (column [0:2] - both included) present the voltage supplied to each of the three piezos. 
#Column [3:5] contain the measured intensities by the three photo diodes. Column 6 contain the sensor temperature multiplied by 100. Column [7:9] contain integers which are 
#used to calculate the actual temperature of the sensor. The final three columns (if present) display the GSK voltage, the GFID voltage and Gain respectively. 
#Only the first 6 columns are recorded during initial learnings and all other values are represented by np.nan.

def import_output(path):
	temp = []
	with open(path, 'r') as file:
	    lines = file.readlines()
	    df = np.zeros([len(lines),len(lines[-1].split())])*np.nan
	    for i in range(len(lines)):
	        splits = lines[i].split()
	        for j in range(len(splits)):
	            df[i,j] = splits[j]
	    section_last_idx = np.where(np.diff(df[:,0])<0)
	    sections = [df[0:section_last_idx[0][0]+1]]
	    for i in range(len(section_last_idx[0])-1):
	        sections.append(df[section_last_idx[0][i]+1:section_last_idx[0][i+1]+1])
	    sections.append(df[section_last_idx[0][-1]+1:])
	    return sections
    # return 2+2


import numpy as np

class tmm_stack():
    def __init__(self, lam = 10e-6):
        lam0 = 10.5e-6 #central wavelength - used for calculation of thin-film thickness.
        self.temperature = 273.15+20 #The temperature is used in calculation of Ge refractive index
        self.custom_indices = False
        self.refrac_funcs = []
        self.refrac_funcs.append(self.refrac_Air)
        self.refrac_funcs.append(self.refrac_Ge)
        # self.refrac_funcs.append(self.refrac_ZnS)
        self.refrac_funcs.append(self.refrac_ThF4)
        # self.refrac_funcs.append(self.refrac_ZnS)
        self.refrac_funcs.append(self.refrac_Ge)
        self.refrac_funcs.append(self.refrac_ZnSe)
        self.indices = np.zeros(len(self.refrac_funcs), dtype = np.complex_)
        self.update_indices(lam0)
        self.thicknesses = np.array([np.inf, 0.25*lam0/self.indices[1].real, 0.25*lam0/self.indices[2].real, 0.5*lam0/self.indices[3].real, np.inf])
        self.materials = ['Air', 'Ge', 'ZnS', 'ThF4', 'ZnS', 'Ge', 'ZnSe'] #Refrective index in first column, thickness in the second
        # self.transfer_matrix = np.zeros([2,1], dtype = np.complex_)
        self.transfer_matrix = self.calculate_transfer_matrix()
        self.lam = lam

    #This function updates the refractive indices to match the inpu vacuum wavelength, lam. This also updates the system global wavelength, If no input is given, then nothing changes
    def update_indices(self, lam = None):
        if self.custom_indices == False:
            if lam != None:
                self.lam = lam
            for i in range(len(self.refrac_funcs)):
                self.indices[i] = self.refrac_funcs[i](self.lam)

    #This function calculates the system transfer matrix using Transfer Matrix Method. 
    def calculate_transfer_matrix(self, lam = None):
        if lam == None:
            lam = self.lam
        else:
            self.lam = lam
            self.update_indices(lam)
        self.transfer_matrix = self.interface(self.indices[0], self.indices[1]) @ self.ac_phase(self.indices[1],self.thicknesses[1], lam)
        for i in np.arange(2, len(self.thicknesses[:])-1):
            self.transfer_matrix = self.transfer_matrix @ self.interface(self.indices[i-1], self.indices[i]) @ self.ac_phase(self.indices[i], self.thicknesses[i], lam)
        self.transfer_matrix = self.transfer_matrix @ self.interface(self.indices[-2], self.indices[-1])
        return self.transfer_matrix

    #This function calculates the transmittance of the stack at the supplied wavelength. If no wavelength is given, then the transmittance is calculated for the current system wide wavelength. 
    def transmittance(self, lam = None):
        if lam == None:
            lam = self.lam
            M = self.transfer_matrix
        else: 
            M = self.calculate_transfer_matrix(lam)
        return abs(1/M[0,0])**2

    #This function calculates the reflectance of the stack at the supplied wavelength. If no wavelength is given, then the reflectance is calculated for the current system wide wavelength. 
    def reflectance(self, lam = None):
        if lam == None:
            lam = self.lam
            M = self.transfer_matrix
        else: 
            M = self.calculate_transfer_matrix(lam)
        return np.abs(M[1,0]/M[0,0])**2

    #This function calculates (1-transmittance-reflectance) of the stack at the supplied wavelength. If no wavelength is given, then the difference is calculated for the current system wide wavelength. 
    def loss(self, lam = None):
        if lam == None:
            lam = self.lam
            M = self.transfer_matrix
        else: 
            M = self.calculate_transfer_matrix(lam)
        return 1 - abs(1/M[0,0])**2 - np.abs(M[1,0]/M[0,0])**2

    #This funcion calculates the refractive index of ThF4 at the supplied wavelength. If no wavelength is passed on, the function uses the system wide wavelength.
    def refrac_ThF4(self, lam = None):
        # Handbook of Optical Constants of Solids II, 1998, page 1051
        if lam == None:
            lam = self.lam
        lam = lam * 1e6
        return 1.118 + 2.46 / lam - 2.98 / (lam**2) - 0.0j

    #This funcion calculates the refractive index of ZnSe at the supplied wavelength. If no wavelength is passed on, the function uses the system wide wavelength.
    def refrac_ZnSe(self, lam = None):
        if lam == None:
            lam = self.lam 
        # Handbook of Optical Constants of Solids II, page 737
        lam = lam * 1e6
        return np.sqrt(4 + 1.9*lam**2/(lam**2 - 0.113)) - 0.0j

    #This funcion calculates the refractive index of Ge at the supplied wavelength. If no wavelength is passed on, the function uses the system wide wavelength.
    def refrac_Ge(self, lam = None):
        if lam == None:
            lam = self.lam
        # Handbook of Optical Constants of Solids I, page 465
        lam = lam * 1e6
        A = -6.04e-3 * self.temperature + 11.05128
        B = 9.295e-3 * self.temperature + 4.00536
        C = -5.392e-4 * self.temperature + 0.599034
        D = 4.151e-4 * self.temperature + 0.09145
        E = 1.51408 * self.temperature + 3426.5
        return np.sqrt(A + (B * lam**2) / (lam**2 - C) + (D * lam**2) / (lam**2 - E)) - 0.0j

    #This funcion calculates the refractive index of ZnS at the supplied wavelength. If no wavelength is passed on, the function uses the system wide wavelength.
    def refrac_ZnS(self, lam = None):
        if lam == None:
            lam = self.lam
        # https://refractiveindex.info/?shelf=main&book=ZnS&page=Debenham
        lam = lam * 1e6
        return np.sqrt(8.393 + 0.14383/(lam**2 - 0.2421**2) + 4430.99/(lam**2 - 36.71**2)) - 0.0j

    #This funcion calculates the refractive index of air at the supplied wavelength. If no wavelength is passed on, the function uses the system wide wavelength.
    def refrac_Air(self, lam = None):
        if lam == None:
            lam = self.lam
        return 1.0 - 0.0j

    #This function is used to calculate the light behaviour at a material interface. n1 is refractive index of leftmost film, while n2 is refracive index of rightmost film
    def interface(self, n1, n2):
        D = np.array([[0+0j,0+0j], [0+0j,0+0j]])
        D[0, 0] = 1 + n2 / n1
        D[0, 1] = 1 - n2 / n1
        D[1, 0] = 1 - n2 / n1
        D[1, 1] = 1 + n2 / n1
        D = 0.5 * D
        return D

    #This function calculates the accumulated phase as the light travels a distance d through a material with refractive index n. lam is the vacuum wavelength.
    def ac_phase(self, n, d, lam = None):
        if lam == None:
            lam = self.lam
        P = np.array([[0+0j,0+0j], [0+0j,0+0j]])
        P[0,0] = np.exp(1j * n * 2 * np.pi * d / lam)
        P[1, 1] = np.exp(-1j * n * 2 * np.pi * d / lam)
        return P

    #This function flips the list of layers in the stack. This should not make a difference to the plots, but might be usefull in other circumstances 
    def flip_stack(self):
        self.indices = np.flip(self.indices)
        self.thicknesses = np.flip(self.thicknesses)
        self.materials.reverse()
        self.refrac_funcs.reverse()

    #This function calculates the transfer matrix at a specific point inside the stack
    def transfer_matrix_at_x(self, x, lam = None):
        if lam == None:
            lam = self.lam
        self.calculate_transfer_matrix(lam)
        if (x > np.sum(self.thicknesses[1:-1])) | (x < 0): #make sure x is within the stack
            return np.nan
        else:
            temp = 0
            layer_nb = 1 #starting at 1 since layer 0 (and the last as well) is infinite
            for i in np.arange(1, len(self.thicknesses)-1): #Count which layer x is in
                temp += self.thicknesses[i]
                if temp >= x:
                    layer_nb = i
                    break

        d = np.sum(self.thicknesses[1:layer_nb+1]) - x #remaining distance of that layer
        
        #Calculate transfer matrix for the rest of the stack to the right of x
        if layer_nb < len(self.indices[:])-2:
            M = self.ac_phase(self.indices[layer_nb], d, lam) @ self.interface(self.indices[layer_nb],self.indices[layer_nb+1])
            for i in np.arange(layer_nb+1, len(self.indices)-2):
                M = M @ self.ac_phase(self.indices[i], self.thicknesses[i], lam) @ self.interface(self.indices[i], self.indices[i+1])
            M = M @ self.ac_phase(self.indices[-2], self.thicknesses[-2], lam) @ self.interface(self.indices[-2], self.indices[-1])
        else:
            M = self.ac_phase(self.indices[layer_nb], d, lam) @ self.interface(self.indices[layer_nb], self.indices[-1])
        return M

    #This function calculates the electric field of the rightward and leftward traveling wave at a point, x, inside the stack. 
    def E_at_x(self, x, lam = None):
        M = self.transfer_matrix_at_x(x, lam)
        
        E0 = np.ones([2,1], dtype = np.complex_) #Electric field to the left of the stack
        EN = np.zeros([2,1], dtype = np.complex_) #Electric field to the right of the stack
        EN[0,0] = E0[0,0]/self.transfer_matrix[0,0]
        E0[1,0] = self.transfer_matrix[1,0]*EN[0,0]

        E = np.ones([2,1], dtype = np.complex_)
        E[0,0] = M[0,0]*EN[0,0] #Rightwards traveling E-field
        E[1,0] = M[1,0]*EN[0,0] #Leftwards traveling E-field
        return E

    #This function can be used to set the system wide wavelength of the system. When this is done, the refracrtive indices and transfer matrix is also updated.
    def set_wavelength(self, lam):
        self.lam = lam
        self.update_indices(lam)
        _ = self.calculate_transfer_matrix(lam)

    #This function can be used to pass a 1D numpy array of layer thicknesses. 
    def set_thicknesses(self, array_1D):
        self.thicknesses = array_1D

    #This function can be used to pass a list of strings containing names of each material in the stack. 
    def set_materials(self, list_of_strings):
        self.materials = list_of_strings

    #This function can be used to manually set the refractive indices of the stack. This is passed in as a numpy array. 
    def set_indices(self, array_1D):
        self.custom_indices = True
        self.indices = array_1D

    #This function is used to add additional refractive index functions. The function name is simply passed to this function which then appends it to a list. The functions should only take 
    #a single input (wavelength in meters) and give a single output (complex refractive index.)
    def add_refractive_funcs(self, function):
        self.refrac_funcs.append(function)

    #The germanium layers have temperature dependent refrective index. This function sets the system wide temperature. 
    def set_temperature(self, temperature):
        self.temperature = temperature

    #This function clears all information about the stack and resets it. It must then be built from scratch.
    def clear_stack(self):
        self.refrac_funcs = []
        self.indices = np.array([])
        self.thicknesses = np.array([])
        self.materials = []
        self.custom_indices = False



import numpy as np

def dsc2np(file_location):

	with open(file_location, 'r') as file:
	    lines = file.readlines()
	    
	    #Find positions of where the header ends and the footer begins. 
	    #Then load these into their own lists
	    header_idx = 0
	    footer_idx = 0

	    #The header ends where the data starts:
	    while lines[header_idx] != '1) DSC 8500 Temperature Scan \n':
	        header_idx += 1
	    header = lines[0:header_idx]
	    
	    #The first line of the footer
	    while lines[footer_idx] != '\tDSC8500 MANUAL TUNE CALIBRATION VALUES: \n':
	        footer_idx += 1
	    footer = lines[footer_idx:]
	    
	    #The steps in the recipe are compiled into its own list
	    run_steps = []
	    #Find position of first step
	    for i in range(len(header)):
	        if '1)' in header[i]:
	            first_step_idx = i
	    #Find the remaining steps
	    for i in range(first_step_idx,len(header)-1):
	        if ')' in header[i]:
	            run_steps.append(header[i])
	            
	    #The index of the recipe steps inside the main data block 
	    run_step_idx = []
	    j = 0
	    for i in range(header_idx,footer_idx):
	        if run_steps[j].split("\t")[0] in lines[i]:
	            run_step_idx.append(i)
	            if j < len(run_steps)-1:
	                j += 1
	    
	    #Labels of the columns of the output matrices
	    explination = ['Time [min]', 'Unsubtracted Heat Flow [mW]', 'Baseline Heat Flow [mW]', 'Program Temperature [°C]',\
	                  'Sample Temperature [°C]', 'Approx. Gas Flow [-]', 'Heat FLow Calibration [-]', 'Uncorrected Heat Flow [mW]']
	    
	    #A numpy array is generated for each recipe step (entry in run_steps)
	    numpy_blocks = []
	    for i in range(len(run_steps)):
	        if i == 0:
	            block = lines[run_step_idx[i]+3 : run_step_idx[i+1]]
	        elif i == len(run_steps) -1:
	            block = lines[run_step_idx[i]+1 : footer_idx-2]
	        else:
	            block = lines[run_step_idx[i]+1 : run_step_idx[i+1]]
	        
	        stepsweep = np.zeros([len(block),8])
	        for j in range(stepsweep.shape[0]):
	            for k, substring in enumerate(block[j].split("\t")):
	                if substring != '' and substring != ' \n':
	                    if substring == '-1.#IND00' or substring == '-1.#IND00 \n':
	                        stepsweep[j,k-1] = np.inf
	                    else:
	                        stepsweep[j,k-1] = float(substring)
	        numpy_blocks.append(stepsweep)

	return numpy_blocks, header, footer, run_steps, explination


from . import animate_data_cube
from . import basic_math
from . import black_body_spectrum
from . import density_scatter_plot
from . import dsc2numpy
from . import dsc_class
from . import fill_gaps
from . import fpi_sim_matrix_angular
from . import fpi_sim_matrix
from . import fpi_sim
from . import FPI_TMM
from . import FPI_class
from . import hsti_export
from . import hsti_import
from . import import_camera_response
from . import import_newtec_cm
from . import import_mirror_sep
from . import import_output
from . import least_squares_methods
from . import lossy_fpi_sim_matrix
from . import mirror_sep2wavelength
from . import mse
from . import PCA_class
from . import r_sq
from . import remove_bad_px
from . import remove_vignetting
from . import repair_image_defects
from . import save_band
from . import spectral_debending
from . import voronoi_partitioning

from .animate_data_cube import animation
from .basic_math import mean_center, autoscale, norm_normalization, msc, normalize, \
normalize_cube, subtract_band, flatten, inflate, median_filter_cube, \
targeted_median_filter, load_cube, hottelings, conf95lim, array2rgb, apply_NUC_cube, \
apply_NUC_image, naive_temperature_image, naive_temperature_cube, relative_mirror_separation,\
correct_laser_pixels, correct_laser_pixels_large
from .black_body_spectrum import bb_frac_lam, bb_exitance_lam, bb_frac_k, bb_exitance_k
from .density_scatter_plot import density_scatter_plot
from .dsc2numpy import dsc2np
from .fill_gaps import remove_zeros, avg_neighbors
from .fpi_sim_matrix_angular import FPI_trans_matrix_ang
from .fpi_sim_matrix import FPI_trans_matrix
from .fpi_sim import FPI_trans
from .hsti_export import export_data_cube
from .import_newtec_cm import import_cm
from .hsti_import import import_data_cube, import_image_acquisition_settings
from .import_camera_response import import_camera_response
from .import_mirror_sep import import_mirror_sep
from .import_output import import_output
from .lossy_fpi_sim_matrix import FPI_trans_matrix_lossy
from .mirror_sep2wavelength import ms2wl
from .mse import lst_mse
from .r_sq import r_sq
from .remove_vignetting import remove_vignette
from .repair_image_defects import repair_bad_col, repair_nans, remove_stuck_px, remove_outlying_px
from .save_band import band2img
from .spectral_debending import debend
from .voronoi_partitioning import fps, voronoi


## Class files
from .PCA_class import PCA
from .dsc_class import dsc
from .least_squares_methods import ALS,GLS
from .FPI_TMM import Stack
from .FPI_class import FPI



import numpy as np

# This function calculates the fraction of the energy contained below wavelength, lam, compared to the total energy of a black body emitter at a given temperature. 
# This is done without numerical integration using the method proposed in  https://www.ijee.ie/articles/Vol20-6/IJEE1557.pdf
def bb_frac_lam(lam, temperature):
    h = 6.62607004e-34 #[m2 kg / s]
    c = 299792458 #[m / s]
    kB = 1.380649e-23 #[J/K]
    C2 = h*c/kB
    z = C2/(lam*temperature)
    N = 5
    F = 0
    for i in np.arange(1,N+1):
        ci = z**3/i + 3*z**2/(i**2) + 6*z/(i**3) + 6/(i**4)
        F += ci*np.exp(-i*z)
    return 15*F/(np.pi**4)

# This function uses bb_frac_lam to calculate a spectrum of exitances. The input wavelengths, lams, are the center wavelengths of each 'bin' within wich the exitance is calculated.
# The black body temperature is also supplied. 
def bb_exitance_lam(lams, temperature):
    #lams are the central wavelengths of each 'bin'. A new wavelength axis is generated, where all points are
    #shifted half a wavelength step to form the intervals in which the black body fractions are calculated
    temp_lams = np.zeros(len(lams)+1)
    dlams = np.diff(lams)
    temp_lams[1:-1] = lams[:-1] + dlams/2
    temp_lams[0] = lams[0] - dlams[0]/2
    temp_lams[-1] = lams[-1] + dlams[-1]/2

    bb_fracs = np.zeros_like(temp_lams)
    for i in range(len(temp_lams)):
        bb_fracs[i] = bb_frac_lam(temp_lams[i], temperature)
        
    h = 6.62607004e-34 #[m2 kg / s]
    c = 299792458 #[m / s]
    kB = 1.380649e-23 #[J/K]
    sigma = 2*np.pi**5*kB**4/(15*c**2*h**3)
    return abs(np.diff(bb_fracs))*(sigma*temperature**4) #[W/m^2] for each bin. Smaller bins of course lead to smaller exitance in given interval


# The following functions are the same as previously, but rewritten to work with wavenumbers instead. 

def bb_frac_k(k, temperature):
    h = 6.62607004e-34 #[m2 kg / s]
    c = 299792458 #[m / s]
    kB = 1.380649e-23 #[J/K]
    C2 = h*c/kB
    C1 = h*c**2
    z = k*C2/T
    N = 5
    F = 0
    for n in np.arange(1,N+1):
        F += - np.exp(-n*z) * (z**3/n + 3*z**2/(n**2) + 6*z/(n**3) + 6/(n**4)) + 6/(n**4)
    return F*2*np.pi*C1/(sigma*C2**4)

def bb_exitance_k(wvnbs, temperature):
    #wvnbs are the central wavenumber of each 'bin'. A new wavenumber axis is generated, where all points are
    #shifted half a wavenumber step to form the intervals in which the black body fractions are calculated
    temp_wvnbs = np.zeros(len(wvnbs)+1)
    dwvnbs = np.diff(wvnbs)
    temp_wvnbs[1:-1] = wvnbs[:-1] + dwvnbs/2
    temp_wvnbs[0] = wvnbs[0] - dwvnbs[0]/2
    temp_wvnbs[-1] = wvnbs[-1] + dwvnbs[-1]/2

    bb_fracs = np.zeros_like(temp_wvnbs)
    for i in range(len(temp_wvnbs)):
        bb_fracs[i] = bb_frac_k(temp_wvnbs[i], temperature)
        
    h = 6.62607004e-34 #[m2 kg / s]
    c = 299792458 #[m / s]
    kB = 1.380649e-23 #[J/K]
    sigma = 2*np.pi**5*kB**4/(15*c**2*h**3)
    return abs(np.diff(bb_fracs))*(sigma*temperature**4) #[W/m^2] for each bin. Smaller bins of course lead to smaller exitance in given interval



import numpy as np

class Stack():
    def __init__(self, airgap, lam):
        self.lam0 = 10.5e-6 #central wavelength - used for calculation of thin-film thickness.
        self.temperature = 273.15+20 #The temperature is used in calculation of Ge refractive index
        self.gap = airgap #Size of the airgap between bragg mirror and gold mirror
        self.layers = np.array([self.refrac_Ge(self.lam0, self.temperature), self.refrac_ThF4(self.lam0), self.refrac_Ge(self.lam0, self.temperature), 1.0+0.0j,\
        self.refrac_Ge(self.lam0, self.temperature), self.refrac_ThF4(self.lam0), self.refrac_Ge(self.lam0, self.temperature)])
        self.layers = np.c_[self.layers, np.array([0.5*self.lam0/self.layers[0].real, 0.25*self.lam0/self.layers[1].real,  0.25*self.lam0/self.layers[2].real, self.gap,\
        0.25*self.lam0/self.layers[4].real, 0.25*self.lam0/self.layers[5].real, 0.5*self.lam0/self.layers[6].real])]
        self.layer_material = ['ZnSe', 'Ge', 'ThF4', 'Ge', 'Air', 'Ge', 'ThF4', 'Ge', 'ZnSe'] #Refrective index in first column, thickness in the second
        # self.layers = np.array([self.refrac_Ge(self.lam0, self.temperature), self.refrac_ThF4(self.lam0), self.refrac_Ge(self.lam0, self.temperature), 1])
        # self.layers = np.c_[self.layers, np.array([0.5*self.lam0/self.layers[0], 0.25*self.lam0/self.layers[1],  0.25*self.lam0/self.layers[2], self.gap])] #Refrective index in first column, thickness in the second
        self.transfer_matrix = np.zeros([2,1], dtype = np.complex_)
        self.lam = lam

    def update_indices(self, lam):
        self.layers[:,0] = np.array([self.refrac_Ge(lam, self.temperature), self.refrac_ThF4(lam), self.refrac_Ge(lam, self.temperature), 1.0+0.0j,\
        self.refrac_Ge(lam, self.temperature), self.refrac_ThF4(lam), self.refrac_Ge(lam, self.temperature)])
        # self.layers[:,0] = np.array([self.refrac_Ge(lam, self.temperature), self.refrac_ThF4(lam), self.refrac_Ge(lam, self.temperature), 1])

    def set_airgap(self, gap):
        self.layers[3,1] = gap
        self.stack_transfer_matrix(self.lam)

    def set_wavelength(self, lam):
        self.lam = lam
        self.stack_transfer_matrix(self.lam)

    def stack_transfer_matrix(self, lam):
        self.update_indices(lam)
        self.transfer_matrix = self.interface(self.refrac_ZnSe(lam), self.layers[0,0]) @ self.ac_phase(self.layers[0,0],self.layers[0,1].real, lam)
        for i in np.arange(1, len(self.layers[:,0]), 1):
            self.transfer_matrix = self.transfer_matrix @ self.interface(self.layers[i-1,0], self.layers[i,0]) @ self.ac_phase(self.layers[i,0], self.layers[i,1].real, lam)
        self.transfer_matrix = self.transfer_matrix @ self.interface(self.layers[-1,0], self.refrac_ZnSe(lam))
        # self.transfer_matrix = self.transfer_matrix @ np.ones([2,2])*1e8 #assume that the final layer is totally reflective. Ideally, the matrix would contain infs
        return self.transfer_matrix

    def transmittance(self, lam):
        M = self.stack_transfer_matrix(lam)
        return abs(1/M[0,0])**2

    def reflectance(self, lam):
        M = self.stack_transfer_matrix(lam)
        return np.abs(M[1,0]/M[0,0])**2

    def loss(self, lam):
        M = self.stack_transfer_matrix(lam)
        return 1 - abs(1/M[0,0])**2 - np.abs(M[1,0]/M[0,0])**2

    def refrac_ThF4(self, lam):
        # Handbook of Optical Constants of Solids II, 1998, page 1051
        lam = lam * 1e6
        n = 1.118 + 2.46 / lam - 2.98 / (lam**2) - 1.0j
        return n

    def refrac_ZnSe(self, lam):
        # Handbook of Optical Constants of Solids II, page 737
        lam = lam * 1e6
        n = np.sqrt(4 + 1.9*lam**2/(lam**2 - 0.113)) - 0.0j
        return n

    def refrac_Ge(self, lam, temp):
        # Handbook of Optical Constants of Solids I, page 465
        lam = lam * 1e6
        A = -6.04e-3 * temp + 11.05128
        B = 9.295e-3 * temp + 4.00536
        C = -5.392e-4 * temp + 0.599034
        D = 4.151e-4 * temp + 0.09145
        E = 1.51408 * temp + 3426.5
        n = np.sqrt(A + (B * lam**2) / (lam**2 - C) + (D * lam**2) / (lam**2 - E)) - 0.0j
        return n

    def interface(self, n1, n2):
        #n1 is refractive index of leftmost film, while n2 is refracive index of rightmost film
        D = np.array([[0+0j,0+0j], [0+0j,0+0j]])
        D[0, 0] = 1 + n2 / n1
        D[0, 1] = 1 - n2 / n1
        D[1, 0] = 1 - n2 / n1
        D[1, 1] = 1 + n2 / n1
        D = 0.5 * D
        return D

    def ac_phase(self, n, d, lam):
        P = np.array([[0+0j,0+0j], [0+0j,0+0j]])
        P[0,0] = np.exp(1j * n * 2 * np.pi * d / lam)
        P[1, 1] = np.exp(-1j * n * 2 * np.pi * d / lam)
        return P

    def E_at_x(self, lam, x):
        self.stack_transfer_matrix(lam)
        if x > np.sum(self.layers[:,1]):
            return np.nan
        else:
            temp = 0
            layer_nb = 0
            for i in range(len(self.layers[:,1])):
                temp += self.layers[i,1].real
                if temp >= x:
                    layer_nb = i
                    break

        d = np.sum(self.layers[0:layer_nb+1,1].real) - x #remaining distance of that layer

        if layer_nb < len(self.layers[:,0])-1:
            M = self.ac_phase(self.layers[layer_nb,0], d, lam) @ self.interface(self.layers[layer_nb,0],self.layers[layer_nb+1,0])
            for i in np.arange(layer_nb+1, len(self.layers[:,0])-1):
                M = M @ self.ac_phase(self.layers[i,0], self.layers[i,1].real, lam) @ self.interface(self.layers[i,0], self.layers[i+1,0])
            M = M @ self.ac_phase(self.layers[-1,0], self.layers[-1,1].real, lam) @ self.interface(self.layers[-1,0], self.refrac_ZnSe(lam))
            # M = M @ self.ac_phase(self.layers[-1,0], self.layers[-1,1].real, lam) @ np.ones([2,2])*1e8
        else:
            M = self.ac_phase(self.layers[layer_nb,0], d, lam) @ self.interface(self.layers[layer_nb,0],self.refrac_ZnSe(lam))
            # M = self.ac_phase(self.layers[layer_nb,0], d, lam) @ np.ones([2,2])*1e8

        E0 = np.ones([2,1], dtype = np.complex_)
        EN = np.zeros([2,1], dtype = np.complex_)
        EN[0,0] = E0[0,0]/self.transfer_matrix[0,0]
        E0[1,0] = self.transfer_matrix[1,0]*EN[0,0]

        E = np.ones([2,1], dtype = np.complex_)
        E[0,0] = M[0,0]*EN[0,0]
        E[1,0] = M[1,0]*EN[0,0]
        return E



import numpy as np
from HSTI import fpi_sim_matrix
import sys
from scipy import interpolate as sci_inter
import string
import pickle

def ms2wl(mirror_seps, lams, camera_response, mirror_sep_spec, correction_factor=1.5e4):

    trans_matrix = fpi_sim_matrix.FPI_trans_matrix(mirror_seps, lams, 293)

    if type(camera_response) == str:
        sys.modules['scipy.interpolate._interpolate'] = sci_inter
        with open(camera_response, 'rb') as f:
            interpolate_response = pickle.load(f)
        del sys.modules['scipy.interpolate._interpolate']
        camera_response = interpolate_response(lams)
        trans_matrix = trans_matrix*camera_response[np.newaxis,:]
    elif (type(camera_response) == np.ndarray) and (len(camera_response) == len(lams)):
        camera_response = camera_response
        trans_matrix = trans_matrix*camera_response[np.newaxis,:]
    else:
        print('Camera response bust either be a string pointing to a .pkl file containing \
        the calibration curve, or be a numpy vector with the same length as the wavelength axis.')

    correction_matrix = np.zeros([trans_matrix.shape[1], trans_matrix.shape[1]])
    for j in range(trans_matrix.shape[1]):
        correction_matrix[j,j] = 2
        if j != 0:
            correction_matrix[j,j-1] = -1
        if j != trans_matrix.shape[1]-1:
            correction_matrix[j,j+1] = -1

    lams_diff = np.diff(lams)
    lams_diff = np.append(lams_diff, lams_diff[-1])
    trans_matrix = trans_matrix*lams_diff[np.newaxis,:]
    trans_matrix = trans_matrix/np.max(trans_matrix)

    ws_spec = np.linalg.inv(trans_matrix.T@trans_matrix + correction_factor*correction_matrix)@trans_matrix.T@mirror_sep_spec
    ws_spec = ws_spec/np.max(ws_spec)
    return ws_spec


import cv2 as cv
import numpy as np
import os
from natsort import natsorted

def import_data_cube(directory, file_format = '.ppm', rotate = True, verbose = False):
    """
    ########## import_data_cube ##########
This function imports images of the stated file type (.ppm by default) from the
provided directory. The functions first looks for images at the provided distination,
but if none is found it looks first for a folder named 'capture' containing
images, but if that also fails it looks for directory 'images/capture'. If all
fails, an error message is displayed. All images are sorted based on their names
using the natsorted algorithm in the natsort module before they are compiled
into a 3D numpy array with the spatial dimensions given in the first two axes,
and the spectral dimension along the third axes.
The function also prints out the dimensions of the data cube if the verbose
parameter is set.
    """

    try:
        if (os.path.exists(directory)) and (sum([ file_format in s for s in os.listdir(directory)]) > 0):
            new_directory = directory
        elif (os.path.exists(f"{directory}/capture/")) and (sum([ file_format in s for s in os.listdir(f"{directory}/capture/")]) > 0):
            new_directory = f"{directory}/capture/"
        elif (os.path.exists(f"{directory}/images/capture/")) and (sum([ file_format in s for s in os.listdir(f"{directory}/images/capture/")]) > 0):
            new_directory = f"{directory}/images/capture/"

        file_lst = [elem for elem in os.listdir(new_directory) if file_format in elem]
        file_lst = natsorted(file_lst)
        imgs = []
        if rotate:
            for img in file_lst:
                imgs.append(np.rot90(cv.imread(f'{new_directory}{img}',cv.IMREAD_ANYDEPTH)))
        else:
            for img in file_lst:
                imgs.append(cv.imread(f'{new_directory}{img}',cv.IMREAD_ANYDEPTH))
        imgs = np.array(imgs, dtype = 'float64')
        imgs = np.moveaxis(imgs, 0, 2)
        if verbose:
            print('Hyperspectral image shape:')
            print(imgs.shape)

    except:
        print('Chosen directory either does not exist or contains no images of provided file type')

    return imgs

    """
    ########## import_image_acquisition_settings ##########
    his function imports the image acquisition settings during the capturing event.
    The directory that the function uses as input must be the one containing the
    'images' directory.
    """
def import_image_acquisition_settings(directory, verbose = False):

    with open(f"{directory}/output.txt", 'r') as file:
        lines = file.readlines()
        temperature_lst = []
        for line in lines:
            if (len(line.split()) == 13) or (len(line.split()) == 10):
                temperature_lst.append(int(line.split()[6])/1000)
        sens_T = np.mean(temperature_lst)
        if len(lines[-1].split()) == 13:
            GSK = int(lines[-1].split()[10])
            GFID = int(lines[-1].split()[11])
            Gain = float(lines[-1].split()[12])
        else:
            GSK = None
            GFID = None
            Gain = None

    if verbose:
        print(f'Sensor temperature: {sens_T}')
        print(f'GSK: {GSK}')
        print(f'GFID: {GFID}')
        print(f'Gain: {round(Gain,2)}')

    valdict = {
      'SENS_T': sens_T,
      'GSK': GSK,
      'GFID': GFID,
      'GAIN': Gain
    }

    return valdict



from scipy import interpolate as sci_inter
import pickle
import sys

################################################################################
################################################################################
#### Function for importing camera response saved in .pkl calibration file. ####
### The output is a scipy interp1d object which must then be supplied with  ####
################# wavelengths to generate the camera response ##################
################################################################################
################################################################################

def import_camera_response(path):
    if type(path) == str:
        sys.modules['scipy.interpolate._interpolate'] = sci_inter
        with open(path, 'rb') as f:
            interpolate_response = pickle.load(f)
        del sys.modules['scipy.interpolate._interpolate']
    else:
        print('Camera response must be a string pointing to a .pkl file containing the calibration curve.')
    return interpolate_response



import numpy as np
import os
import concurrent.futures as cf
import cv2

# Function for saving individual images in .ppm format
def save_img(img, directory):
    cv2.imwrite(directory, img)
    # Renaming the '.pgm' images to '.ppm' to follow the format coming fron the camera
    new_directory = directory[0:-2] + 'p' + directory[-2+1:]
    os.rename(directory, new_directory)

def export_data_cube(data_cube, directory, rotate = False, file_name_list = None, generate_folders = False, normalize_each_band = False):
    """
    ########## HSTI_export ##########
    This function takes an HSTI numpy array and exports it as individual .ppm
    images to a folder given by folder_name.
    """
    if generate_folders:
        if os.path.isdir(os.getcwd() + '/' + directory) == False:
            os.mkdir(os.getcwd() + '/' + directory)
            os.mkdir(os.getcwd() + '/' + directory + '/images')
            os.mkdir(os.getcwd() + '/' + directory + '/images/capture')
        new_directory = os.getcwd() + '/' + directory + '/images/capture'
    else:
        if os.path.isdir(os.getcwd() + '/' + directory) == False:
            os.mkdir(os.getcwd() + '/' + directory)
        new_directory = os.getcwd() + '/' + directory
    if rotate:
        data_cube = np.rot90(data_cube, 3)
    img_size = [data_cube.shape[0], data_cube.shape[1]]

    if normalize_each_band:
        for i in range(data_cube.shape[2]):
            data_cube[:,:,i] = 65535.9*(data_cube[:,:,i] - np.nanmin(data_cube[:,:,i]))/(np.nanmax(data_cube[:,:,i]) - np.nanmin(data_cube[:,:,i]))
    else:
        data_cube = 65535.9*(data_cube - np.nanmin(data_cube))/(np.nanmax(data_cube) - np.nanmin(data_cube))

    data_cube = data_cube.astype(np.uint16)


# setting up for multi processing
    img_lst, directory_lst = [], []
    if (file_name_list == None) or (len(file_name_list) != data_cube.shape[2]):
        for i in range(data_cube.shape[2]):
            img_lst.append(data_cube[:,:,i])
            step = i*10
            directory_lst.append(f'{new_directory}/step{step}.pgm')
    else:
        for i in range(data_cube.shape[2]):
            img_lst.append(data_cube[:,:,i])
            directory_lst.append(f'{new_directory}/{file_name_list[i]}.pgm')

# Multiprocessing step
    with cf.ProcessPoolExecutor() as executor:
        results = executor.map(save_img, img_lst, directory_lst)



import numpy as np
import sys
from HSTI import fpi_sim_matrix
import os
import string
from scipy import interpolate as sci_inter
import pickle

################################################################################
################################################################################
## Class for performing FPI calculations such as converting from FTIR to FPI ###
## spectra or vice versa. All is based on the assumption of loss less mirrors ##
################################################################################
################################################################################


class FPI:
    def __init__(self, mirror_seps = [], lams = [], temperature = None):
        self.trans_matrix = []
        self.__original_trans_matrix = []
        self.mirror_seps = mirror_seps
        self.lams = lams
        self.temperature = temperature
        self.camera_response = []
        if (self.lams != []) and (self.mirror_seps != []) and (self.temperature != None):
            self.trans_matrix = fpi_sim_matrix.FPI_trans_matrix(self.mirror_seps, self.lams, self.temperature)

    def set_trans_matrix(self, trans_matrix):
        self.trans_matrix = trans_matrix

    def set_mirror_seps(self, mirror_seps):
        self.mirror_seps = mirror_seps

    def set_wavelengths(self, lams):
        self.lams = lams

    def get_trans_matrix(self):
        return self.trans_matrix

    def get_mirror_seps(self):
        return self.mirror_seps

    def get_wavelengths(self):
        return self.lams

    def wl2ms(self, wavelength_spec):
        if self.trans_matrix != []:
            lams_diff = np.diff(self.lams)
            lams_diff = np.append(lams_diff, lams_diff[-1])
            temp_trans_matrix = np.copy(self.trans_matrix)
            temp_trans_matrix = temp_trans_matrix*lams_diff

            ms_spec = temp_trans_matrix@wavelength_spec
            return ms_spec
        else:
            print('A transmission matrix must first be calculated or supplied')

    def ms2wl(self, mirror_sep_spec, correction_factor=1.5e4):
        correction_matrix = np.zeros([self.trans_matrix.shape[1], self.trans_matrix.shape[1]])
        for j in range(self.trans_matrix.shape[1]):
            correction_matrix[j,j] = 2
            if j != 0:
                correction_matrix[j,j-1] = -1
            if j != self.trans_matrix.shape[1]-1:
                correction_matrix[j,j+1] = -1

        lams_diff = np.diff(self.lams)
        lams_diff = np.append(lams_diff, lams_diff[-1])
        temp_trans_matrix = np.copy(self.trans_matrix)
        temp_trans_matrix = temp_trans_matrix*lams_diff[np.newaxis,:]
        temp_trans_matrix = temp_trans_matrix/np.max(temp_trans_matrix)

        ws_spec = np.linalg.inv(temp_trans_matrix.T@temp_trans_matrix + correction_factor*correction_matrix)@temp_trans_matrix.T@mirror_sep_spec
        ws_spec = ws_spec/np.max(ws_spec)
        return ws_spec

    def apply_camera_response(self, camera_response = []):
        if (self.trans_matrix != []) and (self.__original_trans_matrix == []):
            self.__original_trans_matrix = np.copy(self.trans_matrix)

            if type(camera_response) == str:
                sys.modules['scipy.interpolate._interpolate'] = sci_inter
                with open(camera_response, 'rb') as f:
                    interpolate_response = pickle.load(f)
                del sys.modules['scipy.interpolate._interpolate']
                self.camera_response = interpolate_response(self.lams)
                self.trans_matrix = self.__original_trans_matrix*self.camera_response
            elif (type(camera_response) == np.ndarray) and (len(camera_response) == len(self.lams)):
                self.camera_response = camera_response
                self.trans_matrix = self.__original_trans_matrix*self.camera_response
            else:
                print('Camera response bust either be a string pointing to a .pkl file containing \
                the calibration curve, or be a numpy vector with the same length as the wavelength axis.')
        else:
            print('A transmission matrix must first be calculated or supplied')

    def remove_camera_response(self):
        if self.__original_trans_matrix != []:
            self.trans_matrix = self.__original_trans_matrix
            self.__original_trans_matrix = []
        else:
            print('The camera response was not applied to the transmission matrix in the first place. Nothing has been done')



import numpy as np

def remove_stuck_px(data_cube):
    temp_data_cube = np.copy(data_cube)
    min_img = np.min(data_cube, axis=2)
    zero_idx = np.where(min_img == 0)
    zero_std_idx = np.where(np.std(data_cube, axis = 2) == 0)
    for i in range(len(zero_idx[0])):
        if zero_idx[0][i] == 0 and zero_idx[1][i] == 0:  # upper left
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :]) / 3
        elif zero_idx[0][i] == 0 and zero_idx[1][i] == data_cube.shape[1]-1:  # upper right
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 3
        elif zero_idx[0][i] == data_cube.shape[0]-1 and zero_idx[1][i] == 0:  # lower left
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] - 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :]) / 3
        elif zero_idx[0][i] == data_cube.shape[0]-1 and zero_idx[1][i] == data_cube.shape[1]-1:  # lower right
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] - 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 3
        elif zero_idx[0][i] == 0:  # upper row
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 5
        elif zero_idx[0][i] == data_cube.shape[0]-1:  # lower row
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] - 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 5
        elif zero_idx[1][i] == 0:  # leftmost column
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :]) / 5
        elif zero_idx[1][i] == data_cube.shape[1]-1:  # rightmost column
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] - 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 5
        else:
            temp_data_cube[zero_idx[0][i], zero_idx[1][i], :] = (data_cube[zero_idx[0][i] + 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i] + 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i] - 1, zero_idx[1][i], :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_idx[0][i], zero_idx[1][i] - 1, :]) / 8

    for i in range(len(zero_std_idx[0])):
        if zero_std_idx[0][i] == 0 and zero_std_idx[1][i] == 0:  # upper left
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :]) / 3
        elif zero_std_idx[0][i] == 0 and zero_std_idx[1][i] == data_cube.shape[1]-1:  # upper right
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 3
        elif zero_std_idx[0][i] == data_cube.shape[0]-1 and zero_std_idx[1][i] == 0:  # lower left
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :]) / 3
        elif zero_std_idx[0][i] == data_cube.shape[0]-1 and zero_std_idx[1][i] == data_cube.shape[1]-1:  # lower right
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 3
        elif zero_std_idx[0][i] == 0:  # upper row
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 5
        elif zero_std_idx[0][i] == data_cube.shape[0]-1:  # lower row
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 5
        elif zero_std_idx[1][i] == 0:  # leftmost column
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :]) / 5
        elif zero_std_idx[1][i] == data_cube.shape[1]-1:  # rightmost column
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 5
        else:
            temp_data_cube[zero_std_idx[0][i], zero_std_idx[1][i], :] = (data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i] - 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i] + 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i] - 1, zero_std_idx[1][i], :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] + 1, :] + \
                                                                 data_cube[zero_std_idx[0][i], zero_std_idx[1][i] - 1, :]) / 8

    return temp_data_cube


def remove_outlying_px(data_cube, cut_off):
    temp_data_cube = np.copy(data_cube)
    gray_img = np.mean(temp_data_cube, axis=2)
    gray_img += abs(np.min(gray_img))
    gray_img = gray_img / np.max(gray_img)

    neighbour_img = np.zeros([gray_img.shape[0], gray_img.shape[1]])
    for i in range(gray_img.shape[0]):
        for j in range(gray_img.shape[1]):
            if i == 0 and j == 0:  # upper left
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j + 1] + gray_img[i + 1, j] + gray_img[i + 1, j + 1]) / 3
            elif i == 0 and j == gray_img.shape[1] - 1:  # upper right
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j - 1] + gray_img[i + 1, j] + gray_img[i + 1, j - 1]) / 3
            elif i == gray_img.shape[0] - 1 and j == 0:  # bottom left
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i - 1, j] + gray_img[i, j + 1] + gray_img[i - 1, j + 1]) / 3
            elif i == gray_img.shape[0] - 1 and j == gray_img.shape[1] - 1:  # bottom right
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j - 1] + gray_img[i - 1, j] + gray_img[i - 1, j - 1]) / 3
            elif i == 0:  # top row
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j - 1] + gray_img[i, j + 1] + gray_img[i + 1, j] + gray_img[i + 1, j - 1] + gray_img[i + 1, j + 1]) / 5
            elif i == gray_img.shape[0] - 1:  # bottom row
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i, j - 1] + gray_img[i, j + 1] + gray_img[i - 1, j] + gray_img[i - 1, j - 1] + gray_img[i - 1, j + 1]) / 5
            elif j == 0:  # leftmost column
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i + 1, j] + gray_img[i - 1, j] + gray_img[i, j + 1] + gray_img[i + 1, j + 1] + gray_img[i - 1, j + 1]) / 5
            elif j == gray_img.shape[1] - 1:  # rightmost column
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i + 1, j] + gray_img[i - 1, j] + gray_img[i, j - 1] + gray_img[i + 1, j - 1] + gray_img[i - 1, j - 1]) / 5
            else:
                neighbour_img[i, j] = gray_img[i, j] - (gray_img[i - 1, j - 1] + gray_img[i - 1, j] + gray_img[i - 1, j + 1] + gray_img[i, j + 1] + gray_img[i + 1, j + 1] + gray_img[i + 1, j] + gray_img[i + 1, j - 1] + gray_img[i, j - 1]) / 8

    outlier_idx = np.where(abs(neighbour_img) > cut_off)

    for i in range(len(outlier_idx[0])):
        if outlier_idx[0][i] == 0 and outlier_idx[1][i] == 0:  # upper left
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :]) / 3
        elif outlier_idx[0][i] == 0 and outlier_idx[1][i] == data_cube.shape[1]-1:  # upper right
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 3
        elif outlier_idx[0][i] == data_cube.shape[0]-1 and outlier_idx[1][i] == 0:  # lower left
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :]) / 3
        elif outlier_idx[0][i] == data_cube.shape[0]-1 and outlier_idx[1][i] == data_cube.shape[1]-1:  # lower right
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 3
        elif outlier_idx[0][i] == 0:  # upper row
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 5
        elif outlier_idx[0][i] == data_cube.shape[0]-1:  # lower row
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 5
        elif outlier_idx[1][i] == 0:  # leftmost column
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :]) / 5
        elif outlier_idx[1][i] == data_cube.shape[1]-1:  # rightmost column
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 5
        else:
            temp_data_cube[outlier_idx[0][i], outlier_idx[1][i], :] = (data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i] - 1, :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i] + 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i] - 1, outlier_idx[1][i], :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] + 1, :] + \
                                                                 data_cube[outlier_idx[0][i], outlier_idx[1][i] - 1, :]) / 8

    return temp_data_cube, outlier_idx



import numpy as np
from IPython.display import clear_output

#Function which distributes n_seeds (a numper of points) equally within a lists
#of points to obtain furthest point sampling.

#The function takes in a list of points. Every entry in the list contains both the
#x and y coordinate of a given point. It returns the coordinates of the selected
#sample points
def fps(points, n_seeds):
    """
    https://minibatchai.com/ai/2021/08/07/FPS.html
    points: [N, 3] array containing the whole point cloud
    n_seeds: samples you want in the sampled point cloud typically << N
    """
    points = np.array(points)

    # Represent the points by their indices in points
    points_left = np.arange(len(points)) # [P]

    # Initialise an array for the sampled indices
    sample_inds = np.zeros(n_seeds, dtype='int') # [S]

    # Initialise distances to inf
    dists = np.ones_like(points_left) * float('inf') # [P]

    # Select a point from points by its index, save it
    selected = 0
    sample_inds[0] = points_left[selected]

    # Delete selected
    points_left = np.delete(points_left, selected) # [P - 1]

    # Iteratively select points for a maximum of n_seeds
    for i in range(1, n_seeds):
        # Find the distance to the last added point in selected
        # and all the others
        last_added = sample_inds[i-1]

        dist_to_last_added_point = (
            (points[last_added] - points[points_left])**2).sum(-1) # [P - i]

        # If closer, updated distances
        dists[points_left] = np.minimum(dist_to_last_added_point,
                                        dists[points_left]) # [P - i]

        # We want to pick the one that has the largest nearest neighbour
        # distance to the sampled points
        selected = np.argmax(dists[points_left])
        sample_inds[i] = points_left[selected]

        # Update points_left
        points_left = np.delete(points_left, selected)

    return points[sample_inds]


def voronoi(array_2D, n_seeds):
    p_x = np.arange(0,array_2D.shape[1])
    p_y = np.arange(0,array_2D.shape[0])
    points = np.zeros([len(p_x)*len(p_y),2])
    c = 0
    for i in p_x:
        for j in p_y:
            points[c, 0] = i
            points[c, 1] = j
            c += 1
    domains = []
    for i in range(n_seeds):
        domains.append([])

    temp = fps(points, n_seeds)
    seeds = []
    for i in range(len(temp)):
        # The selected seed points are converted for faster calculations later
        seeds.append(temp[i,0] + temp[i,1]*1j)
    #The same is done for the rest of the points_left

    points = list(points[:,0] +  points[:,1]*1j)

    for i in range(len(points)):
        if i%1000 == 0:
            clear_output(wait=True)
            print('Partitioning array: ' + str(round(100*(i+1)/len(points), 1)) + '%')
        distance = []
        for j in range(n_seeds):
            distance.append(abs(points[i]-seeds[j]))
        shortest_dist = np.where(distance == np.min(distance))
        domains[shortest_dist[0][0]].append(points[i])

    clear_output(wait=True)
    print('Partitioning image: 100%')

    return domains, seeds



import numpy as np
from scipy.ndimage import median_filter
from IPython.display import clear_output
import concurrent.futures as cf
import os
import cv2 as cv
import pkg_resources
from scipy.signal import savgol_filter
from natsort import natsorted
from scipy.signal import find_peaks
from scipy import interpolate


#Multiple of these functions takes an optional argument defining in which axis
#the function operates. If 's' is passed as an argument, the operation is performed
#along the spectral axis (across bands/variables). If the argument is 'b', the
#function operates on each band/variable individually

#Subtrats the mean from the data, either the mean of each spectrum (axis = 's')
#or the mean of each band (axis = 'b')
def mean_center(data_cube, axis = 's'):
    if axis == 's':
        return data_cube - np.nanmean(data_cube, axis = 2)[:,:,np.newaxis]
    elif axis == 'b':
        return data_cube - np.nanmean(data_cube, axis = (0,1))
    else:
        raise Exception("Invalid input argument. Axis input argument must be either 's' or 'b'")

#Subtracts mean and scales with STD - the old standardize function. Setting
#axis = 's' is the same as doing SNV (standard normal variate)
def autoscale(data_cube, axis = 's'):
    if axis == 's':
        std_cube = np.std(data_cube, axis = 2)[:,:,np.newaxis]
        std_cube[std_cube < 1e-6] = np.nan
        return (data_cube - np.mean(data_cube, axis = 2)[:,:,np.newaxis])/std_cube
    elif axis == 'b':
        return (data_cube - np.mean(data_cube, axis = (0,1)))/np.std(data_cube, axis = (0,1))
    else:
        raise Exception("Invalid input argument. Axis input argument must be either 's' or 'b'")

#Uses norm of given order for normalization. If axis = 's', then each spectrum is
#divided by its norm. If axis = 'b', then every band is divided by the norm of
#the entire band.
def norm_normalization(data_cube, order, axis = 's'):
    if axis == 's':
        return data_cube/np.linalg.norm(data_cube, ord = order, axis = 2)[:,:,np.newaxis]
    elif axis == 'b':
        return data_cube/np.linalg.norm(data_cube, ord = order, axis = (0,1))
    else:
        raise Exception("Invalid input argument. Axis input argument must be either 's' or 'b'")

#Apply multiplicative scatter correction to entire datacube. If no reference
#spectrum is supplied, the mean spectrum of the cube is used instead.
def msc(data_cube, ref_spec = None):
    if ref_spec == None:
        ref_spec = np.nanmean(data_cube, axis = (0,1))
        ref_spec_mean = np.nanmean(ref_spec)
        ref_spec = ref_spec - ref_spec_mean
    else:
        ref_spec_mean = np.nanmean(ref_spec)
        ref_spec - ref_spec_mean
    ref_spec = ref_spec[:,np.newaxis]
    X = mean_center(data_cube, 's')
    X = X.reshape([X.shape[0]*X.shape[1], X.shape[2]])
    X = np.rot90(X, 3)

    b = np.linalg.inv(ref_spec.T@ref_spec)@ref_spec.T@X
    b[b<1e-8] = np.nan
    X_new = X/b + ref_spec_mean
    X_new = np.rot90(X_new)
    X_new = X_new.reshape([data_cube.shape[0], data_cube.shape[1], data_cube.shape[2]])
    return X_new

#Setting axis = 's' is the same as normalizing each spectrum (pixel) to span from 0 to 1
#axis = 'b' normalizes each band individually
def normalize(data_cube, axis = 's'):
    if axis == 's':
        cube = data_cube - np.nanmin(data_cube, axis = 2)[:,:,np.newaxis]
        max_cube = np.nanmax(cube, axis = 2)[:,:,np.newaxis]
        return cube/max_cube
    elif axis == 'b':
        cube = data_cube - np.nanmin(data_cube, axis = (0,1))
        return cube/np.nanmax(cube, axis = (0,1))
    else:
        raise Exception("Invalid input argument. Axis input argument must be either 's' or 'b'")

#Normalizes the cube to span from 0 to 1
def normalize_cube(data_cube):
    cube = data_cube - np.nanmin(data_cube)
    return cube/np.nanmax(cube)

#Subtracts selected band from all layers in the cube
def subtract_band(data_cube, band):
    return data_cube - data_cube[:,:,band][:,:,np.newaxis]

#Flattens data cube into 2D array
def flatten(data_cube):
    return np.reshape(data_cube, [data_cube.shape[0]*data_cube.shape[1], data_cube.shape[2]])

#Reshapes flattened data cube back into 3 dimensions
def inflate(array_2D, n_rows, n_cols):
    return np.reshape(array_2D, [n_rows, n_cols, array_2D.shape[1]])

#Median filters each band in data cube
def median_filter_cube(data_cube, kernel_size):
    band_lst = []
    for i in range(data_cube.shape[2]):
        band_lst.append(data_cube[:,:,i])
    kernel_size_lst = np.ones(len(band_lst), dtype = int)*kernel_size
    with cf.ThreadPoolExecutor() as executor:
        results = executor.map(median_filter, band_lst, kernel_size_lst)
    filtered_cube = np.zeros_like(data_cube)
    for i, result in enumerate(results):
        filtered_cube[:,:,i] = result
    return filtered_cube

#Median filters a 2D array and only applies it to the locations marked as True
#in the px_idx array.
def targeted_median_filter(input_array, px_idx, kernel_size):
    if input_array.shape != px_idx.shape:
        print("Arrays must be the same shape. px_idx must be a true/false numpy array")
        return
    if type(input_array[0,0]) is not np.float64:
         input_array = input_array.astype(float)
    filtered_array = median_filter(input_array, size = kernel_size)
    input_array[px_idx] = filtered_array[px_idx]
    return input_array

#Imports cube based on directory argument
def load_cube(directory):
    if '.pam' in directory:
        data = np.fromfile(directory,  dtype=np.uint8)   #importing filename to data
        header = []
        i = 0
        while (header[i-8:i] != ['E','N','D','H','D','R','\r','\n']):
            header.append(chr(data[i]))
            i += 1
        #Make header into a single string and split into list of strings whenever there is a space
        string_lst = ''.join(header).replace('\r\n', ' ').split(' ')
        N_rows = int(string_lst[string_lst.index('WIDTH')+1])
        N_wvls = int(string_lst[string_lst.index('HEIGHT')+1])
        N_cols = int(string_lst[string_lst.index('DEPTH')+1])
        cube = np.reshape(data[i:], [N_rows ,N_cols ,N_wvls], order = 'F')
        cube = cube.astype(float)
    elif '.npy' in directory:
        cube = np.load(directory)
        cube = cube.astype(float)
    else:
        file_list = [x for x in os.listdir(f'{directory}/images/capture') if ".ppm" and not "RGB" in x]
        file_list = natsorted(file_list)
        cube = []
        for file in file_list: #load bands into data structure
            cube.append(np.rot90(cv.imread(f'{directory}/images/capture/{file}',cv.IMREAD_ANYDEPTH)))
        cube = np.array(cube, dtype = 'float64')
        cube = np.moveaxis(cube, 0, 2) #Rearranges the array to required shape
    return cube

#Calculates hottelings T^2 statistic for matrix X, where the variables are
#represented by each column and the samples by the rows
#https://learnche.org/pid/latent-variable-modelling/principal-component-analysis/hotellings-t2-statistic
def hottelings(X):
    return np.nansum((X/np.nanstd(X, axis=0))**2, axis=1)

#Calculates the upper and lower 95% confidence limits of the mean of input vector, x
def conf95lim(x):
    conf = []
    mean = np.nanmean(x)
    std = np.nanstd(x)
    conf.append(mean-2*std)
    conf.append(mean+2*std)
    return conf

#Select three layers (either as a list or numpy vector) and use these three as the
#channels of an rgb image. The first layer is the red channel, the second layer
#the green channel and the third layer is the blue channel.
def array2rgb(data_cube, three_layers):
    three_layer_cube = np.zeros([data_cube.shape[0], data_cube.shape[1], 3])
    three_layer_cube[:,:,0] = data_cube[:,:,three_layers[0]] - np.nanmin(data_cube[:,:,three_layers[0]])
    three_layer_cube[:,:,0] = three_layer_cube[:,:,0]/np.nanmax(three_layer_cube[:,:,0])
    three_layer_cube[:,:,1] = data_cube[:,:,three_layers[1]] - np.nanmin(data_cube[:,:,three_layers[1]])
    three_layer_cube[:,:,1] = three_layer_cube[:,:,1]/np.nanmax(three_layer_cube[:,:,1])
    three_layer_cube[:,:,2] = data_cube[:,:,three_layers[2]] - np.nanmin(data_cube[:,:,three_layers[2]])
    three_layer_cube[:,:,2] = three_layer_cube[:,:,2]/np.nanmax(three_layer_cube[:,:,2])
    return three_layer_cube

#This function calculates and applies a NUC to the entire datacube. The NUC is
#dependent on the sensor temperature and the GSK settings of the camera. The
#NUC is calculated from camera specific calibration files from the accompanying NUC directory
def apply_NUC_cube(cube, sensor_temp, GSK, camera_ID = '10_10_200_191'):

    resource_package = __name__
    resource_path_offset = '/'.join(('HSTI_data_files-main', f"NUC/{camera_ID}/offset_coefs_1st_order.npy"))  # Do not use os.path.join()
    resource_path_slope = '/'.join(('HSTI_data_files-main', f"NUC/{camera_ID}/slope_coefs_1st_order.npy"))  # Do not use os.path.join()
    path_offset = pkg_resources.resource_stream(resource_package, resource_path_offset)
    path_slope = pkg_resources.resource_stream(resource_package, resource_path_slope)
    offset_coefs = np.load(path_offset, allow_pickle = True)
    slope_coefs = np.load(path_slope, allow_pickle = True)

    M = np.array([[GSK, sensor_temp, 1]]).T

    offsets = offset_coefs@M
    slopes = slope_coefs@M

    mean_offset = np.nanmean(offsets)
    mean_slopes = np.nanmean(slopes)

    offset_correction = - mean_slopes*offsets/slopes + mean_offset
    slope_correction = (mean_slopes - slopes)/slopes

    offset_matrix = offset_correction.reshape([cube.shape[0], cube.shape[1]], order = 'C')
    slope_matrix = slope_correction.reshape([cube.shape[0], cube.shape[1]], order = 'C')

    temp_cube = np.copy(cube)
    for i in range(temp_cube.shape[2]):
        temp_cube[:,:,i] = temp_cube[:,:,i] + temp_cube[:,:,i]*slope_matrix + offset_matrix
    return temp_cube

#This function calculates and applies a NUC to a single image. The NUC is dependent
#on the sensor temperature and the GSK settings of the camera. The NUC is
#calculated from camera specific calibration files from the accompanying NUC directory.
def apply_NUC_image(image, sensor_temp, GSK, camera_ID = '10_10_200_191'):

    resource_package = __name__
    resource_path_offset = '/'.join(('HSTI_data_files-main', f"NUC/{camera_ID}/offset_coefs_1st_order.npy"))  # Do not use os.path.join()
    resource_path_slope = '/'.join(('HSTI_data_files-main', f"NUC/{camera_ID}/slope_coefs_1st_order.npy"))  # Do not use os.path.join()
    path_offset = pkg_resources.resource_stream(resource_package, resource_path_offset)
    path_slope = pkg_resources.resource_stream(resource_package, resource_path_slope)
    offset_coefs = np.load(path_offset, allow_pickle = True)
    slope_coefs = np.load(path_slope, allow_pickle = True)


    M = np.array([[GSK, sensor_temp, 1]]).T

    offsets = offset_coefs@M
    slopes = slope_coefs@M

    mean_offset = np.nanmean(offsets)
    mean_slopes = np.nanmean(slopes)

    offset_correction = - mean_slopes*offsets/slopes + mean_offset
    slope_correction = (mean_slopes - slopes)/slopes

    offset_matrix = offset_correction.reshape([image.shape[0], image.shape[1]], order = 'C')
    slope_matrix = slope_correction.reshape([image.shape[0], image.shape[1]], order = 'C')

    return image  + image*slope_matrix + offset_matrix

def naive_temperature_image(image, sensor_temp, GSK, camera_ID = '10_10_200_191'):
    resource_package = __name__
    resource_path_offset = '/'.join(('HSTI_data_files-main', f"Temperature_conversions/{camera_ID}/temp_offset_coefs_1st_order.npy"))  # Do not use os.path.join()
    resource_path_slope = '/'.join(('HSTI_data_files-main', f"Temperature_conversions/{camera_ID}/temp_slope_coefs_1st_order.npy"))  # Do not use os.path.join()
    path_offset = pkg_resources.resource_stream(resource_package, resource_path_offset)
    path_slope = pkg_resources.resource_stream(resource_package, resource_path_slope)
    offset_coefs = np.load(path_offset, allow_pickle = True)
    slope_coefs = np.load(path_slope, allow_pickle = True)

    a = slope_coefs[:,0]*GSK + slope_coefs[:,1]*sensor_temp + slope_coefs[:,2]
    b = offset_coefs[:,0]*GSK + offset_coefs[:,1]*sensor_temp + offset_coefs[:,2]

    flat_img = np.reshape(image, image.size)
    flat_img = flat_img*a + b
    return np.reshape(flat_img, [image.shape[0], image.shape[1]])

def naive_temperature_cube(data_cube, sensor_temp, GSK, camera_ID = '10_10_200_191'):
    resource_package = __name__
    resource_path_offset = '/'.join(('HSTI_data_files-main', f"Temperature_conversions/{camera_ID}/temp_offset_coefs_1st_order.npy"))  # Do not use os.path.join()
    resource_path_slope = '/'.join(('HSTI_data_files-main', f"Temperature_conversions/{camera_ID}/temp_slope_coefs_1st_order.npy"))  # Do not use os.path.join()
    path_offset = pkg_resources.resource_stream(resource_package, resource_path_offset)
    path_slope = pkg_resources.resource_stream(resource_package, resource_path_slope)
    offset_coefs = np.load(path_offset, allow_pickle = True)
    slope_coefs = np.load(path_slope, allow_pickle = True)

    a = slope_coefs[:,0]*GSK + slope_coefs[:,1]*sensor_temp + slope_coefs[:,2]
    b = offset_coefs[:,0]*GSK + offset_coefs[:,1]*sensor_temp + offset_coefs[:,2]

    flat_cube = np.reshape(data_cube, [data_cube.shape[0]*data_cube.shape[1], data_cube.shape[2]])
    flat_cube = flat_cube*a[:,np.newaxis] + b[:,np.newaxis]
    return np.reshape(flat_cube, [data_cube.shape[0], data_cube.shape[1], data_cube.shape[2]])

def rsvd(input_matrix, rank, power_iter = 3): #https://towardsdatascience.com/intuitive-understanding-of-randomized-singular-value-decomposition-9389e27cb9de
    Omega = np.random.randn(input_matrix.shape[1], rank)
    Y = input_matrix @ Omega
    for q in range(power_iter):
        Y = input_matrix @ (input_matrix.T @ Y)
    Q, _ = np.linalg.qr(Y)
    B = Q.T @ input_matrix
    u_tilde, s, v = np.linalg.svd(B, full_matrices = False)
    u = Q @ u_tilde
    return u, s, v

def relative_mirror_separation(directory):

    temp = []
    with open(f'{directory}/output.txt', 'r') as file:
        lines = file.readlines()
        df = np.zeros([len(lines),len(lines[-1].split())])*np.nan
        for i in range(len(lines)):
            splits = lines[i].split()
            for j in range(len(splits)):
                df[i,j] = splits[j]
        section_last_idx = np.where(np.diff(df[:,0])<0)
        sections = [df[0:section_last_idx[0][0]+1]]
        for i in range(len(section_last_idx[0])-1):
            sections.append(df[section_last_idx[0][i]+1:section_last_idx[0][i+1]+1])
        sections.append(df[section_last_idx[0][-1]+1:])

    diode = sections
    I0 = savgol_filter(diode[-1][:,3], 5, 2, 0)
    I0 = I0-np.min(I0)
    I0 = I0/np.max(I0)
    I1 = savgol_filter(diode[-1][:,4], 5, 2, 0)
    I1 = I1-np.min(I1)
    I1 = I1/np.max(I1)
    I2 = savgol_filter(diode[-1][:,5], 5, 2, 0)
    I2 = I2-np.min(I2)
    I2 = I2/np.max(I2)

    steps = np.arange(len(I0))
    peaks_I0_idx, _ = find_peaks(I0, prominence=1e-1)
    troughs_I0_idx, _ = find_peaks(-I0, prominence=1e-1)
    peaks_I1_idx, _ = find_peaks(I1, prominence=1e-1)
    troughs_I1_idx, _ = find_peaks(-I1, prominence=1e-1)
    peaks_I2_idx, _ = find_peaks(I2, prominence=1e-1)
    troughs_I2_idx, _ = find_peaks(-I2, prominence=1e-1)

    idx0 = np.sort(np.concatenate((peaks_I0_idx, troughs_I0_idx)))
    idx1 = np.sort(np.concatenate((peaks_I1_idx, troughs_I1_idx)))
    idx2 = np.sort(np.concatenate((peaks_I2_idx, troughs_I2_idx)))

    window_len = 10
    all_idx_in_range = np.array([])
    counter = 0
    while len(all_idx_in_range) < 3:
        idx0_in_range = idx0[(idx0>=steps[counter]) & (idx0 < steps[counter+window_len])]
        idx1_in_range = idx1[(idx1>=steps[counter]) & (idx1 < steps[counter+window_len])]
        idx2_in_range = idx2[(idx2>=steps[counter]) & (idx2 < steps[counter+window_len])]
        all_idx_in_range = np.concatenate((idx0_in_range, idx1_in_range, idx2_in_range))
        counter += 1
    idx0 = idx0[idx0 >= all_idx_in_range[0]]
    idx1 = idx1[idx1 >= all_idx_in_range[1]]
    idx2 = idx2[idx2 >= all_idx_in_range[2]]

    dx = 686e-9/4 #dx = (lam/cos(theta))/4 - found using bandpass filters

    x0 = np.arange(len(idx0))*dx
    x1 = np.arange(len(idx1))*dx
    x2 = np.arange(len(idx2))*dx

    p0 = np.poly1d(np.polyfit(idx0, x0, 3))
    p1 = np.poly1d(np.polyfit(idx1, x1, 3))
    p2 = np.poly1d(np.polyfit(idx2, x2, 3))

    R_sqr0 = 1 - np.sum((x0 - p0(idx0))**2)/np.sum((x0 - np.mean(p0(idx0)))**2)
    R_sqr1 = 1 - np.sum((x1 - p1(idx1))**2)/np.sum((x1 - np.mean(p1(idx1)))**2)
    R_sqr2 = 1 - np.sum((x2 - p2(idx2))**2)/np.sum((x2 - np.mean(p2(idx2)))**2)

    p_comb = np.poly1d(np.polyfit(idx2, x2, 3))
    p_comb.coef[0] = (p0.coef[0]+p1.coef[0]+p2.coef[0])/3
    p_comb.coef[1] = (p0.coef[1]+p1.coef[1]+p2.coef[1])/3
    p_comb.coef[2] = (p0.coef[2]+p1.coef[2]+p2.coef[2])/3
    p_comb.coef[3] = (p0.coef[3]+p1.coef[3]+p2.coef[3])/3

    x_comb = np.concatenate((x0,x1,x2))
    idx_comb = np.concatenate((idx0,idx1,idx2))
    R_sqr_cmb = 1 - np.sum((x_comb - p_comb(idx_comb))**2)/np.sum((x_comb - np.mean(p_comb(idx_comb)))**2)

    file_list = [x for x in os.listdir(f'{directory}/images/capture') if ".ppm" and not "RGB" in x]
    file_list = natsorted(file_list)
    cube_steps = [int(s.split('step')[1].split('.ppm')[0]) for s in file_list]

    return p_comb(cube_steps)


def correct_laser_pixels(img_cube):
    mask = np.zeros([10, 10], dtype = bool)
    mask[2,4:7] = True
    mask[3,3:8] = True
    mask[4:7,2:8] = True
    mask[7,3:7] = True
    xx, yy = np.meshgrid(np.arange(0, 10), np.arange(0, 10))
    bad_x = xx[~mask]
    bad_y = yy[~mask]
    for i in range(img_cube.shape[2]):
        temp = img_cube[471:481,381:391,i]
        img_cube[471:481,381:391,i] = interpolate.griddata((bad_x, bad_y), temp[~mask].ravel(), (xx, yy), method = 'linear')
    return img_cube

def correct_laser_pixels_large(img_cube):
    mask = np.zeros([16, 16], dtype = bool)
    mask[2:-2,2:-2] = True
    xx, yy = np.meshgrid(np.arange(0, 16), np.arange(0, 16))
    bad_x = xx[~mask]
    bad_y = yy[~mask]
    for i in range(img_cube.shape[2]):
        temp = img_cube[468:484,378:394,i]
        img_cube[468:484,378:394,i] = interpolate.griddata((bad_x, bad_y), temp[~mask].ravel(), (xx, yy), method = 'linear')
    return img_cube



import numpy as np
def import_pam(directory):
	data = np.fromfile(directory,  dtype=np.uint8)   #importing filename to data

	header = []
	i = 0
	while (header[i-8:i] != ['E','N','D','H','D','R','\r','\n']):
	    header.append(chr(data[i]))
	    i += 1
	#Make header into a single string and split into list of strings whenever there is a space
	string_lst = ''.join(header).replace('\r\n', ' ').split(' ') 
	N_rows = int(string_lst[string_lst.index('WIDTH')+1])
	N_wvls = int(string_lst[string_lst.index('HEIGHT')+1])
	N_cols = int(string_lst[string_lst.index('DEPTH')+1])


	return np.reshape(data[i:], [N_rows ,N_cols ,N_wvls], order = 'F')


from BaselineRemoval import BaselineRemoval
from scipy import interpolate
import numpy as np
import os


class dsc:
	def __init__(self, path_to_txt = None):
		self.header = None
		self.footer = None
		self.run_steps = None
		self.explanations = None
		self.data_blocks = None
		self.sample_weight = None
		self.sample_ID = None
		self.stepscan = False
		self.stepscan_idx = None
		self.Nsteps = None
		self.step_start_temp = None
		self.step_size = None
		self.step_rate = None
		self.step_temperatures = None
		self.step_first_idx = None
		self.step_last_idx = None
		self.stepscan_baseline = None
		if (path_to_txt is not None) and (os.path.isfile(path_to_txt)):
			self.import_dsc_data(path_to_txt)

	def import_dsc_data(self, path_to_txt):
		with open(path_to_txt, 'r') as file:
			lines = file.readlines()
			
			#Find positions of where the header ends and the footer begins. 
			#Then load these into their own lists
			header_idx = 0
			footer_idx = 0

			#The header ends where the data starts:
			while lines[header_idx] != '1) DSC 8500 Temperature Scan \n':
				header_idx += 1
			self.header = lines[0:header_idx]
			
			#The first line of the footer
			while lines[footer_idx] != '\tDSC8500 MANUAL TUNE CALIBRATION VALUES: \n':
				footer_idx += 1
			self.footer = lines[footer_idx:]
			
			#Find the sample ID and sample weight in the header
			i = 0
			while 'Sample ID' not in self.header[i]:
				i += 1
			self.sample_ID = self.header[i].split('\t')[-1].rstrip()

			while 'Display Weight' not in self.header[i]:
				i += 1
			self.sample_weight = float(self.header[i].split('\t')[-1])*1e-6

			#The steps in the recipe are compiled into its own list
			self.run_steps = []
			#Find position of first step
			for i in range(len(self.header)):
				if '1)' in self.header[i]:
					first_step_idx = i
			#Find the remaining steps
			for i in range(first_step_idx,len(self.header)-1):
				if ')' in self.header[i]:
					self.run_steps.append(self.header[i])
					
			#The index of the recipe steps inside the main data block 
			run_step_idx = []
			j = 0
			for i in range(header_idx,footer_idx):
				if self.run_steps[j].split("\t")[0] in lines[i]:
					run_step_idx.append(i)
					if j < len(self.run_steps)-1:
						j += 1
			
			#If a step scan is performed find the number of steps and start temperature
			for i in range(len(self.run_steps)):
				if 'StepScan' in self.run_steps[i]:
					self.stepscan = True
					self.stepscan_idx = i
					self.Nsteps = int(self.run_steps[i].split(' ')[-2].split('X')[0])
					self.step_start_temp = float(self.run_steps[i].split('StepScan:')[-1].split('°C')[0])
					self.step_size = float(self.run_steps[i].split('StepScan:')[-1].split('°C')[1].split('to ')[1]) - self.step_start_temp
					self.step_rate = float(self.run_steps[i].split('@')[-1].split('°C')[0])
					self.step_temperatures = np.arange(self.step_start_temp, self.step_start_temp+self.Nsteps+1)

			#Labels of the columns of the output matrices
			self.explanations = ['Time [s]', 'Unsubtracted Heat Flow [W]', 'Baseline Heat Flow [W]', 'Program Temperature [°C]',\
						  'Sample Temperature [°C]', 'Approx. Gas Flow [-]', 'Heat FLow Calibration [-]', 'Uncorrected Heat Flow [W]']
			
			#A numpy array is generated for each recipe step (entry in run_steps)
			self.data_blocks = []
			for i in range(len(self.run_steps)): #Iterates through each data block (run_steps)
				if i == 0:
					block = lines[run_step_idx[i]+3 : run_step_idx[i+1]] #block is a list of strings
				elif i == len(self.run_steps) -1:
					block = lines[run_step_idx[i]+1 : footer_idx-2]
				else:
					block = lines[run_step_idx[i]+1 : run_step_idx[i+1]]
				
				stepsweep = np.zeros([len(block),len(self.explanations)]) 
				for j in range(stepsweep.shape[0]): #Iterate through each row of data block
					for k, substring in enumerate(block[j].split("\t")): #Split row (line) at each "\t" and iterate through each col
						if substring != '' and substring != ' \n': #Only consider non empty and non-"\n" strings
							if substring == '-1.#IND00' or substring == '-1.#IND00 \n': #If #IND00 - insert nan
								stepsweep[j,k-1] = np.nan
							else:
								stepsweep[j,k-1] = float(substring) #Else convert string to float
				
				#stepsweep now contains all data from the given recipe step. Next we replace any instances of np.nan.
				nan_idx = np.where(np.isnan(stepsweep)) # Get all indices of nans
				for j in range(len(nan_idx[0])): #iterate through indices
					k = 1
					k_increase = 1
					while np.isnan(stepsweep[nan_idx[0][j]+k, nan_idx[1][j]]):  #keep increasing k until a non-nan element is encountered
						if nan_idx[0][j]+k >= stepsweep.shape[0]: #if the index + k is larger than the length of the array, look in the other direction
							k_increase = -1
						k += k_increase
					stepsweep[nan_idx[0][j], nan_idx[1][j]] = stepsweep[nan_idx[0][j]+k, nan_idx[1][j]] #replace nan by closest non-nan

				stepsweep[:,0] = stepsweep[:,0]*60 #convert time from minutes to seconds
				stepsweep[:,1:3] = stepsweep[:,1:3]*1e-3 #convert from mW to W
				#stepsweep[:,3:5] = stepsweep[:,3:5]+273.15 #convert from C to K.
				stepsweep[:,-1] = stepsweep[:,-1]*1e-3
				self.data_blocks.append(stepsweep)

			self.header = [line.replace('\t','').rstrip() for line in self.header]
			self.footer = [line.replace('\t','').rstrip() for line in self.footer]
			self.run_steps = [line.replace('\t','').rstrip() for line in self.run_steps]

	def baseline_scan(self, data, degree): #Remove baseline of order, 'degree', from 'data' 
		BLR_obj = BaselineRemoval(data)
		baselined_data=BLR_obj.IModPoly(degree)
		return baselined_data # Should this be negative in order for heat flow exothermic up



	def baseline_stepscan(self):
		self.step_first_idx = []  #First index at which the program temperature is equal to one of the temperatures in step_temperatures
		self.step_last_idx = []   #Last index the program temperature is equal the same step temperature as in self.step_first_idx
		baseline_idx = []  #The index of the local minimum between the self.step_last_idx of previous step and self.step_first_idx of next step. This is used as the baseline of the curve
		self.step_last_idx.append(0)  #Since the starting temperature technically isn't part of a step (rather the base), the first index is added to the list
		##self.step_last_idx contains one additional idx compared to self.step_first_idx and baseline_idx.

		for i in range(len(self.step_temperatures)-1):
			self.step_first_idx.append(np.where(self.data_blocks[self.stepscan_idx][:,3] == self.step_temperatures[i+1])[-1][0])
			self.step_last_idx.append(np.where(self.data_blocks[self.stepscan_idx][:,3] == self.step_temperatures[i+1])[-1][-1])
			segment = self.data_blocks[self.stepscan_idx][self.step_last_idx[i]:self.step_last_idx[i+1],1] # Looking at an entire step
			slope = (segment[-1] - segment[0])/(self.data_blocks[self.stepscan_idx][self.step_last_idx[i+1],0] - self.data_blocks[self.stepscan_idx][self.step_last_idx[i],0]) #The slope of the line connecting the end points of the step
			temp_time = self.data_blocks[self.stepscan_idx][self.step_last_idx[i]:self.step_last_idx[i+1],0] - self.data_blocks[self.stepscan_idx][self.step_last_idx[i+1],0] #make a time axis the length of the step, starting at t = 0
			slope_correction = segment - (temp_time*slope + segment[0]) #Remove the slope from the segment
			baseline_idx.append(self.step_last_idx[i] + np.where(slope_correction == np.nanmin(slope_correction))[-1][0]) #Find the index of the minimum within the slope corrected segment. This is the baseline

		#remove the base line from the curve and perform linear interpolation at every time step 
		baseline_interp = interpolate.interp1d(self.data_blocks[self.stepscan_idx][baseline_idx,0], self.data_blocks[self.stepscan_idx][baseline_idx,1], kind = 'linear', fill_value = 'extrapolate')
		self.stepscan_baseline = self.data_blocks[self.stepscan_idx][:,1] - baseline_interp(self.data_blocks[self.stepscan_idx][:,0]) #subtract the baseline from every data point in the step scan
		return self.stepscan_baseline

		#Calculate specific heat capacity
	def calculate_cp(self):
		step_baseline_corrected = self.baseline_stepscan() #The step scan must have its baseline removed. 
		areas = []	#Calculate area of each heating step in the step scan.
		for i in range(len(self.step_temperatures)-1):
			areas.append(np.trapz(step_baseline_corrected[self.step_last_idx[i]:self.step_last_idx[i+1]], self.data_blocks[-1][self.step_last_idx[i]:self.step_last_idx[i+1],0]))
		return np.array(areas)/(self.sample_weight*self.step_size)




# Function returns coefficient of determination between values based on model
#(y_fit) and corresponding measured values (y_meas)

def r_sq(y_fit, y_meas):
    mean = sum(y_meas)/len(y_meas)
    #Sum of squared residuals
    SS_res = sum([(m - f)**2 for m, f in zip(y_meas, y_fit)])
    #Total sum of squares
    SS_tot = sum([(m - mean)**2 for m in y_meas])

    R_sq = 1 - SS_res/SS_tot

    return R_sq



import numpy as np
from scipy import interpolate
import concurrent.futures as cf

def debend_single_band(selections, img, interp_spec, mirror_sep):
    interp_img = np.zeros_like(img)
    interp_img[selections[0]] = img[selections[0]]
    for i in range(len(selections) - 1):
        interp_img[selections[i+1]] = interp_spec[i+1](mirror_sep)
    return interp_img

def debend(cube, central_mirror_sep):
    N = 100
    f_length = 35e-3 #Focal length
    px_pitch = 17e-6 #Pixel pitch
    p_x = np.arange(cube.shape[1])
    p_y = np.arange(cube.shape[0])
    P_x, P_y = np.meshgrid(p_x, p_y)
    r = np.sqrt(((P_x - 430)**2 + (P_y - 505)**2)*px_pitch**2)
    theta = np.arctan(r/f_length)
    #Fitting parameters for use later
    [a, b] = [-1.24875814e-04,  2.37220232e-05]

    masks = []
    #THETA hold the angles of each boundary between each ring
    THETA = np.linspace(0, np.max(theta), N+1)
    THETA = np.delete(THETA, 0)
    #THETA = 0 is deleted because this will result in a circle with an area of 0
    # The innermost 'ring' is just a circle and this mask is therefore a special case
    masks.append(theta <= THETA[0])
    #The rest of the masks are generated
    for i in range(N-1):
        masks.append((theta > THETA[i]) & (theta <=THETA[i+1]))

    mirror_sep_axes = []
    mirror_sep_axes.append(central_mirror_sep)
    for i in range(N-1):
        correction = central_mirror_sep*(a * np.rad2deg(THETA[i+1])**2 + b * np.rad2deg(THETA[i+1]))
        mirror_sep_axes.append(central_mirror_sep + correction)

    interpolated_cube = np.zeros_like(cube)
    interpolated_spectra = []
    for i in range(N):
        interpolated_spectra.append(interpolate.interp1d(mirror_sep_axes[i], cube[masks[i], :], fill_value = 'extrapolate'))

    mask_lst, img_lst, interp_spec_lst, mirror_sep_lst = [], [], [], []
    for i in range(cube.shape[2]):
        mask_lst.append(masks)
        img_lst.append(cube[:,:,i])
        interp_spec_lst.append(interpolated_spectra)
        mirror_sep_lst.append(mirror_sep_axes[0][i])

    with cf.ThreadPoolExecutor() as executor:
        results = executor.map(debend_single_band, mask_lst, img_lst, interp_spec_lst, mirror_sep_lst)

    debent = np.zeros_like(cube)
    for i, result in enumerate(results):
        debent[:,:,i] = result

    return debent



import numpy as np
from PIL import Image

def band2img(band, filename):
	# Scale the array to the 0-255 range
	scaled_image = ((band - np.min(band)) / (np.max(band) - np.min(band))) * 255

	# Convert the array to 8-bit unsigned integer type
	uint8_image = scaled_image.astype(np.uint8)

	# Save the array as an 8-bit TIFF file
	im = Image.fromarray(uint8_image)
	im.save(filename)
	return


import numpy as np

class PCA:
    def __init__(self, array_2D = None):
        self.loadings = 0
        self.scores = 0
        self.singular_values = 0
        self.expl_var_ratio = 0
        if ((array_2D is not None) and (len(array_2D.shape) == 2)):
            self.calculate_pca(array_2D)

    #Calculates the PCA of a matrix with each column representing a variable, and the samples represented by the rows.
    def calculate_pca(self, array_2D):
        nan_map = np.isnan(array_2D) #form boolean array with true values everywhere a np.nan is encountered
        if nan_map.any(): #If array_2D contains np.nan, do this
            nan_row_idx = np.unique(np.argwhere(nan_map)[:,0]) #Find every row containing nan
            for idx in nan_row_idx:
                nan_map[idx,:] = True #Expand the map to mark entire row
            cov = np.cov((array_2D[~nan_map].reshape(array_2D.shape[0]-len(nan_row_idx), -1)).T) #exclude marked rows from cov calculation
            del nan_row_idx
        else:
            cov = np.cov(array_2D.T)
        del nan_map
        self.singular_values, self.loadings = np.linalg.eig(cov)
        self.singular_values = np.real(self.singular_values)
        self.loadings = np.real(self.loadings)
        index = self.singular_values.argsort()[::-1] #make sure the singular values are in decending order as well as the loadings
        self.singular_values = self.singular_values[index]
        self.loadings = self.loadings[:,index] #sort the loadings the same way as the singular values
        self.expl_var_ratio = self.singular_values/np.sum(self.singular_values) #calculate the percentage of explained variance for each of the eigen values - this corresponds to the contribution of each PC
        self.scores = array_2D @ self.loadings

    def apply_pca(self, array_2D):
        return array_2D @ self.loadings



from matplotlib.animation import FuncAnimation
from IPython import display
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
from mpl_toolkits.axes_grid1 import make_axes_locatable

def animation(data_cube, title_list, fps, DPI, fixed_color = True, colormap = 'plasma'):
    fig = plt.figure(dpi=DPI)
    fig.set_figheight(8) #set height of the entire figure
    fig.set_figwidth(8) #set width of the entire figure
    gs = gridspec.GridSpec(1, 1) #set size ratio of the subfigures
    ax1 = fig.add_subplot(gs[0,0])
    a=np.zeros((data_cube.shape[0],data_cube.shape[1]))
    if not fixed_color:
        im=plt.imshow(a,interpolation='none', vmin = np.min(data_cube), vmax = np.max(data_cube), cmap = colormap)
    else:
        im=plt.imshow(a,interpolation='none', vmin = np.min(data_cube[:,:,0]), vmax = np.max(data_cube[:,:,0]), cmap = colormap)
    divider = make_axes_locatable(ax1)
    cax = divider.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im, cax=cax, orientation='vertical')
    def animate(frame):
        #update plot
        ax1.set_title(title_list[frame])
        im.set_data(data_cube[:,:,frame])
        if not fixed_color:
            im.set_clim(vmin = np.min(data_cube[:,:,frame]), vmax = np.max(data_cube[:,:,frame]))

    I = 1000/fps
    anim = FuncAnimation(fig, animate, frames=data_cube.shape[2], interval=I)
    video = anim.to_html5_video()
    html = display.HTML(video)
    display.display(html)
    plt.close('all')
    return
